/*
	Downgrade Control -> decrypt.c -> Responsible for decryption patches and hooks
	by Davee
	
	31/12/2010
*/

#include <pspkernel.h>
#include <pspsysmem_kernel.h>

#include <stdio.h>
#include <string.h>
#include <systemctrl.h>

#include "decrypt.h"
#include "patch_table.h"

int pspDecryptPRX(void *exec, u32 exec_size, u32 *out_size);

int (* sceMesgLed_driver_81F72B1F)(void *exec, u32 size, u32 *outsize) = NULL;

int sceResmgr_9DC14891_patched(void *data, u32 datasize, u32 *outsize)
{
	/* call the original function to decrypt the index.dat */
	int res = sceResmgr_driver_9DC14891(data, datasize, outsize);
	
	/* the default version */
	u32 fw_ver = 0x000;
	
	/* now modify the firmware version to our own version */
	switch (sceKernelGetModel())
	{
		/* PSP Phat */
		case 0:
		{
			/* should probably check for other stuff such as TA-82 motherboards etc, fw 330, as updater does specfic stuff for 300ish and lower */
			fw_ver = 0x330;
			break;
		}
		
		/* PSP Slim */
		case 1:
		{
			/* SVEN! */
			fw_ver = 0x360;
			break;
		}
		
		/* PSP 3000 */
		case 2:
		{
			/* minimum firmware is 4.20 */
			fw_ver = 0x420;
			break;
		}
		
		/* PSP 4000 */
		case 3:
		{
			/* minimum firmware is 5.70 */
			fw_ver = 0x570;
			break;
		}
		
		/* PSPgo */
		case 4:
		{
			/* minimum firmware is 5.70 */
			fw_ver = 0x570;
			break;
		}
	}
	
	/* if there is a firmware version, we will generate a version.txt to downgrade */
	if (fw_ver)
	{
		sprintf( data,
				"release:%01X.%01X%01X:\n"
				"build:0000,0,3,1,0:builder@vsh-build6\n"
				"system:56422@release_%03X,0x%08X:\n"
				"vsh:p6576@release_%03X,v57929@release_%03X,20100625:\n"
				"target:1:WorldWide\n",
				/* first line, version (X.YZ) */ (fw_ver >> 8) & 0xF, (fw_ver >> 4) & 0xF, fw_ver & 0xF,
				/* third line, release_XYZ + devkit */ fw_ver, (((fw_ver >> 8) & 0xF) << 24) | (((fw_ver >> 4) & 0xF) << 16) | ((fw_ver & 0xF) << 8) | 0x10,
				/* forth line, release_XYZ + release_XYZ */ fw_ver, fw_ver
				);
	}
	
	/* return the decryption result */
	return res;
}

void PatchMesgled(u32 text_addr)
{
	/* if firmware is >= 6.30 */
	if (sceKernelDevkitVersion() >= 0x06030010)
	{
		/* Patch mesgled to allow older updaters to boot */
		_sw(0, text_addr + g_patch_table.new_updater_check[sceKernelGetModel()]);
	}
}

int sceMesgLed_driver_81F72B1F_patched(void *exec, u32 exec_size, u32 *out_size)
{
	/* do sony decryption */
	int res = sceMesgLed_driver_81F72B1F(exec, exec_size, out_size);
	
	/* check for error */
	if (res != 0)
	{
		/* lets decrypt */
		res = pspDecryptPRX(exec, exec_size, out_size);
	}
	
	/* return result */
	return res;
}

int memlmd_7CF1CD3E_patched(void *exec, u32 exec_size, u32 *out_size)
{
	/* do sony decryption */
	int res = memlmd_7CF1CD3E(exec, exec_size, out_size);
	
	/* check for error */
	if (res != 0)
	{
		/* lets decrypt */
		res = pspDecryptPRX(exec, exec_size, out_size);
	}
	
	/* return result */
	return res;
}
